@@ -4,8 +4,8 @@
 import numpy as np
 import pandas as pd
 
-from pandas import DataFrame, Period, Series
-from datetime import date, timedelta
+from pandas import DataFrame, Period, Series, Timestamp
+from datetime import date
 
 from .decorators import check_attributes
 from .exceptions import CohortAnalysisException
@@ -30,6 +30,7 @@ class CohortMetrics:
         self,
         clickstream: Optional[DataFrame] = None,
         cohorts: Optional[DataFrame] = None,
+        timezone: str = "UTC",
     ):
 
         """
@@ -37,11 +38,13 @@ class CohortMetrics:
 
         Sets the .clickstream and .cohorts metrics (either as DataFrames or None)
 
-        Sets the .filter_logic property
+        Sets the .filter_logic, .timezone, and .current_timestamp properties
 
         """
 
         self.filter_logic = "AND"
+        self.timezone = timezone
+        self.current_timestamp = Timestamp.now(tz=timezone)
 
         self.clickstream = None
         self.cohorts = None
@@ -49,16 +52,12 @@ class CohortMetrics:
         if clickstream is not None:
             if not isinstance(clickstream, DataFrame):
                 raise CohortAnalysisException("clickstream must be a DataFrame if set")
-
-            self.clickstream = clickstream.copy()
-            self.verify_clickstream()
+            self.set_clickstream(clickstream)
 
         if cohorts is not None:
             if not isinstance(cohorts, DataFrame):
                 raise CohortAnalysisException("cohorts must be a DataFrame if set")
-
-            self.cohorts = cohorts.copy()
-            self.verify_cohorts()
+            self.set_cohorts(cohorts)
 
     def __repr__(self):
 
@@ -78,6 +77,8 @@ class CohortMetrics:
         no_clickstream = self.clickstream is None
         no_cohorts = self.cohorts is None
 
+        # TO-DO: indicate the timezone and current_timestamp
+
         if no_clickstream and no_cohorts:
             return f"{classname}(<no clickstream set>, <no cohorts set>)"
         if no_clickstream:
@@ -181,6 +182,90 @@ class CohortMetrics:
                 "cohorts must have reference_timestamp column"
             )
 
+    def change_current_timestamp(
+        self, current_timestamp: Union[str, Timestamp] = "now"
+    ):
+
+        """
+        change_current_timestamp
+
+        By default, CohortMetrics uses the timestamp at initialisation
+        (in the chosen timezone) as its .current_timestamp property.
+
+        This method allows us to change that (e.g. in case working with
+        old/checkpointed data, test data with future values etc)
+
+        KEYWORDS:
+            current_timestamp = "now" - set the new .current_timestamp
+                                        property
+
+            options:
+                "now" - set to current timestamp in the set timezone
+                "last" - set to last timestamp/reference timestamp
+                         contained within the clickstream/cohorts data
+                <any pandas Timestamp> - will be localized in or
+                                         converted to the set timezone
+                <any parseable datetime string> - will be converted to a
+                                                  localized Timestamp
+
+        """
+
+        if current_timestamp == "now":
+
+            self.current_timestamp = Timestamp.now(tz=self.timezone)
+
+        elif current_timestamp == "last":
+
+            clickstream, cohorts = self.filter_by_cohorts()
+
+            self.current_timestamp = max(
+                [clickstream.timestamp.max(), cohorts.reference_timestamp.max()]
+            )
+
+        elif isinstance(current_timestamp, Timestamp):
+
+            if current_timestamp.tz is None:
+                self.current_timestamp = current_timestamp.tz_localize(self.timezone)
+            else:
+                self.current_timestamp = current_timestamp.tz_convert(self.timezone)
+
+        else:
+            self.current_timestamp = Timestamp(current_timestamp, tz=self.timezone)
+
+    def change_timezone(self, timezone="UTC"):
+
+        """
+        change_timezone
+
+        This method changes the .timezone property and converts
+
+            .current_timestamp property
+            .cohorts.reference_timestamp column (if set)
+            .clickstream.timestamp column (if set)
+
+        to the new timezone
+
+        KEYWORDS:
+            timezone = "UTC" - timezone to change instance to
+
+        """
+
+        self.timezone = timezone
+
+        self.current_timestamp = self.current_timestamp.tz_convert(timezone)
+
+        if self.cohorts is not None:
+
+            self.cohorts.loc[
+                :, "reference_timestamp"
+            ] = self.cohorts.reference_timestamp.dt.tz_convert(timezone)
+
+        if self.clickstream is not None:
+
+            self.clickstream.loc[
+                :, "timestamp"
+            ] = self.clickstream.timestamp.dt.tz_convert(timezone)
+
     def change_filter_logic(self, logic: str):
 
         """
@@ -231,24 +316,35 @@ class CohortMetrics:
             If the DataFrame has different names for these columns,
             that can be indicated using these keywords
 
-        # TODO: check the timestamp of clickstream and cohorts having the
-        # same timezone
-        # TODO: set timezone as a class attribute
-        # TODO: raise warnings when any of the data is in the future
-
         """
 
-        self.clickstream = clickstream.copy()
+        clickstream_copy = clickstream.copy()
 
         if timestamp_col != "timestamp":
-            self.clickstream.loc[:, "timestamp"] = clickstream.loc[:, timestamp_col]
+            clickstream_copy.loc[:, "timestamp"] = clickstream.loc[:, timestamp_col]
+        elif "timestamp" not in clickstream.columns:
+            raise CohortAnalysisException(
+                "'timestamp' should be a column of the clickstream DataFrame if timestamp_col not specified"
+            )
 
         if user_id_col != "user_id":
-            self.clickstream.loc[:, "user_id"] = clickstream.loc[:, user_id_col]
+            clickstream_copy.loc[:, "user_id"] = clickstream.loc[:, user_id_col]
+        elif "user_id" not in clickstream.columns:
+            raise CohortAnalysisException(
+                "'user_id' should be a column of the clickstream DataFrame if user_id_col not specified"
+            )
+
+        clickstream_copy.loc[:, "timestamp"] = pd.to_datetime(
+            clickstream_copy.timestamp, utc=True, errors="coerce"  # force to UTC
+        ).dt.tz_convert(
+            self.timezone  # convert to the instance timezone
+        )
+
+        self.clickstream = clickstream_copy
 
         self.verify_clickstream()
 
-        self.clickstream = self.clickstream[["timestamp", "user_id"]]
+        self.clickstream = clickstream_copy[["timestamp", "user_id"]]
 
     def set_cohorts(
         self,
@@ -280,12 +376,26 @@ class CohortMetrics:
         if user_id_col != "user_id":
             cohorts_copy.loc[:, "user_id"] = cohorts.loc[:, user_id_col]
             cohorts_copy = cohorts_copy.drop(user_id_col, axis="columns")
+        elif "user_id" not in cohorts.columns:
+            raise CohortAnalysisException(
+                "'user_id' should be a column of the cohorts DataFrame if user_id_col not specified"
+            )
 
         if reference_timestamp_col != "reference_timestamp":
             cohorts_copy.loc[:, "reference_timestamp"] = cohorts.loc[
                 :, reference_timestamp_col
             ]
             cohorts_copy = cohorts_copy.drop(reference_timestamp_col, axis="columns")
+        elif "reference_timestamp" not in cohorts.columns:
+            raise CohortAnalysisException(
+                "'reference_timestamp' should be a column of the cohorts DataFrame if reference_timestamp_col not specified"
+            )
+
+        cohorts_copy.loc[:, "reference_timestamp"] = pd.to_datetime(
+            cohorts_copy.reference_timestamp, utc=True, errors="coerce"  # force to UTC
+        ).dt.tz_convert(
+            self.timezone  # convert to the instance timezone
+        )
 
         self.cohorts = cohorts_copy
 
@@ -407,7 +517,7 @@ class CohortMetrics:
         if "_MASK" not in columns:
             raise CohortAnalysisException("Passed dataframe should have _MASK column")
 
-        if isinstance(values, str) or (values is None):
+        if isinstance(values, str) or isinstance(values, bool) or (values is None):
             vals = [values]  # wrap single strings into a one-element list
         else:
             vals = values
@@ -503,18 +613,29 @@ class CohortMetrics:
 
         filtered_clickstream, filtered_cohorts = self.__get_attribute_copies()
 
-        # first remove any clicks from the stream with user IDs not in cohorts.user_id
+        # first, remove any users with reference_timestamp later
+        # than current_timestamp property
+
+        filtered_cohorts = filtered_cohorts.loc[
+            filtered_cohorts.reference_timestamp.le(self.current_timestamp)
+        ].drop_duplicates()
+
+        # now remove any clicks from the stream with user IDs not in cohorts.user_id
         filtered_clickstream = filtered_clickstream.merge(filtered_cohorts.user_id)
 
         # next remove any clicks from the stream with timestamps earlier
-        # than the associated user's reference timestamp
+        # than the associated user's reference timestamp and later than
+        # the CohortMetrics instance .current_timestamp property
+
         comparison = filtered_clickstream.join(
             filtered_cohorts.set_index("user_id").reference_timestamp, on="user_id"
         )
 
         not_too_early = comparison.timestamp.ge(comparison.reference_timestamp)
 
-        filtered_clickstream = filtered_clickstream.loc[not_too_early]
+        not_too_late = filtered_clickstream.timestamp.le(self.current_timestamp)
+
+        filtered_clickstream = filtered_clickstream.loc[not_too_early & not_too_late]
 
         # if no further filtering is required, return early:
         if cohort is None:
@@ -589,9 +710,7 @@ class CohortMetrics:
 
         offset_data = data_by_period.copy()
 
-        values_at_end = list(offset_data[target_col].values)
-
-        values_at_start = [0] + values_at_end[:-1]
+        values_at_start = offset_data[target_col].shift(1).fillna(0).astype(int)
 
         offset_data.loc[:, target_col] = values_at_start
 
@@ -1002,15 +1121,13 @@ class CohortMetrics:
         # iterate through clickstream
         dates = pd.date_range(start=min_ts, end=max_ts, normalize=True)
 
-        rolling_dates = [
-            ((day - pd.Timedelta(offset_days, unit="days")), day) for day in dates
-        ]
+        rolling_dates = [(day - pd.DateOffset(days=offset_days), day) for day in dates]
 
         data: Dict[str, List] = {"end_date": [], "n_active": [], "cohort_size": []}
 
         for start, end in rolling_dates:
 
-            after_end = end + pd.Timedelta(1, unit="days")
+            after_end = end + pd.DateOffset(days=1)
 
             mask = clickstream.timestamp.ge(start) & clickstream.timestamp.lt(after_end)
             n_active = clickstream.loc[mask].user_id.nunique()
@@ -1204,9 +1321,8 @@ class CohortMetrics:
 
         return user_activities
 
-    @staticmethod
     def add_interval_upper_bound(
-        cohorts: DataFrame, now_timestamp: pd.Timestamp, interval_in_days: int = 7,
+        self, cohorts: DataFrame, interval_in_days: int = 7
     ) -> DataFrame:
 
         """
@@ -1223,8 +1339,6 @@ class CohortMetrics:
         INPUTS:
             cohorts - filtered cohorts data table
 
-            now_timestamp - timestamp for the current date and time
-
         KEYWORDS:
             interval_in_days = 7 - the number of days for a time interval
 
@@ -1236,7 +1350,7 @@ class CohortMetrics:
 
         cohorts_with_upper_bound = cohorts.assign(
             interval_upper_bound=(
-                now_timestamp - cohorts.reference_timestamp
+                self.current_timestamp - cohorts.reference_timestamp
             ).dt.days.floordiv(interval_in_days)
         )
 
@@ -1314,10 +1428,7 @@ class CohortMetrics:
         return series
 
     def fixed_interval_retention(
-        self,
-        current_timestamp: pd.Timestamp,
-        cohort: CohortFilterType = None,
-        interval_in_days: int = 7,
+        self, cohort: CohortFilterType = None, interval_in_days: int = 7,
     ) -> Series:
 
         """
@@ -1328,11 +1439,6 @@ class CohortMetrics:
         is calculated as "number of active users" divided by "number of
         users having reached this time-interval"
 
-        INPUTS:
-            current_timestamp: pandas Timestamp for retention interval
-                               calculation, ususally set to the timestamp
-                               of now
-
         KEYWORDS:
             cohort = None - allows filtering by cohort. Optional input is
                             a dictionary: keys identify columns, values
@@ -1351,22 +1457,26 @@ class CohortMetrics:
 
         clickstream, cohorts = self.filter_by_cohorts(cohort)
 
-        cohorts = self.add_interval_upper_bound(
-            cohorts, current_timestamp, interval_in_days,
-        )
+        cohorts = self.add_interval_upper_bound(cohorts, interval_in_days)
 
         merged_stream = clickstream.merge(cohorts, on="user_id", how="inner")
 
+        # 1. convert each user's activities to an array of integers
+        ## each integer represents that user is active at that time-interval
+        # 2. count user activities from 0 to N time-interval; 0 for no activity
+        ## N for the largest interval of the cohort
+        # 3. join with cohorts and keep users without any activity
+        # 4. filter user's activity by interval_upper_bound
+        ## activity status can be: active, non-active and impossible to calculate
+        # 5. calculate active rates
+        ## at each time-interval: num of active / num of could be active
+
         cohort_active_rates = (
             self.user_activities_by_interval(merged_stream, interval_in_days)
             .apply(self.fixed_interval_bincount, axis=1)
             .join(cohorts.set_index("user_id"), how="right")
             .apply(self.filter_by_upper_bound, axis=1)
-            .pipe(
-                lambda d: d.sum(axis=0).div(d.notna().sum(axis=0))
-            )  # rate at interval = len(active users) / len(not-na users)
-            # TODO: filter out the data of last time-interval
-            # it's in-completed and not showing proper information
+            .pipe(lambda d: d.sum(axis=0).div(d.notna().sum(axis=0)))
         )
 
         return cohort_active_rates
@@ -1400,10 +1510,7 @@ class CohortMetrics:
         return series
 
     def unbounded_interval_retention(
-        self,
-        current_timestamp: pd.Timestamp,
-        cohort: CohortFilterType = None,
-        interval_in_days: int = 7,
+        self, cohort: CohortFilterType = None, interval_in_days: int = 7,
     ) -> Series:
 
         """
@@ -1415,11 +1522,6 @@ class CohortMetrics:
          of users available". An user is active at the time-interval
         if it is active in a later time-interval.
 
-        INPUTS:
-            current_timestamp: pandas Timestamp for retention interval
-                               calculation, ususally set to the timestamp
-                               of now
-
         KEYWORDS:
             cohort = None - allows filtering by cohort. Optional input is
                             a dictionary: keys identify columns, values
@@ -1437,21 +1539,26 @@ class CohortMetrics:
 
         clickstream, cohorts = self.filter_by_cohorts(cohort)
 
-        cohorts = self.add_interval_upper_bound(
-            cohorts, current_timestamp, interval_in_days,
-        )
+        cohorts = self.add_interval_upper_bound(cohorts, interval_in_days)
 
         merged_stream = clickstream.merge(cohorts, on="user_id", how="inner")
 
+        # 1. convert each user's activities to array of integers
+        ## each integer represents that user is active at that time-interval
+        # 2. count user activities from 0 to N time-interval
+        ## if user is active at K interval, all previous intervals are active
+        # 3. join with cohorts and keep users without any activity
+        # 4. filter user's activity by interval_upper_bound
+        ## activity status can be: active, non-active and impossible to calculate
+        # 5. calculate active rates
+        ## at each time-interval: num of active / num of could be active
+
         cohort_active_rates = (
             self.user_activities_by_interval(merged_stream, interval_in_days)
-            .apply(
-                self.unbounded_interval_bincount, axis=1
-            )  # time-intervals to columns
-            .join(cohorts.set_index("user_id"))
+            .apply(self.unbounded_interval_bincount, axis=1)
+            .join(cohorts.set_index("user_id"), how="right")
             .apply(self.filter_by_upper_bound, axis=1)
-            # calculate active percentage
-            .pipe(lambda d: d.sum(axis=0).div(d.shape[0]))
+            .pipe(lambda d: d.sum(axis=0).div(d.notna().sum(axis=0)))
         )
 
         return cohort_active_rates
@@ -1492,10 +1599,7 @@ class CohortMetrics:
         return series
 
     def wmqy_interval_retention(
-        self,
-        current_timestamp: pd.Timestamp,
-        cohort: CohortFilterType = None,
-        interval_in_days: int = 7,
+        self, cohort: CohortFilterType = None, interval_in_days: int = 7,
     ) -> Series:
 
         """
@@ -1509,10 +1613,6 @@ class CohortMetrics:
         active users in the second to fourth week" divided by "number of
         users in the second to fourth week".
 
-        INPUTS:
-            current_timestamp: pandas Timestamp for retention interval
-                               calculation
-
         KEYWORDS:
             cohort = None - allows filtering by cohort. Optional input is
                             a dictionary: keys identify columns, values
@@ -1556,7 +1656,6 @@ class CohortMetrics:
         interval_type: str = "week",
         melted_output: bool = False,
         retention_cohort_names: List[str] = [],
-        current_timestamp: pd.Timestamp = pd.Timestamp.now(tz=None),
     ):
 
         """
@@ -1587,13 +1686,6 @@ class CohortMetrics:
                                     is the cohort name for the returned
                                     DataFrame
 
-            current_timestamp = pd.Timestamp.now(tz=None) - pandas Timestamp
-                                for the retention metric to calculate the
-                                number of time-intervals, default to the
-                                timestamp when function is invoked
-            # TODO: refactor current_timestamp to be an instance attribute and
-            # calling it in the lower functions
-
         RAISES:
             CohortAnalysisException - if different length for retention_cohort
                                       and retention_cohort_names
@@ -1658,10 +1750,8 @@ class CohortMetrics:
                     name = "/".join([f"{k}={v}" for k, v in cohort.items()])
 
             active_rates = retentions_mapping[retention_type](  # type: ignore
-                current_timestamp, cohort, interval_in_days
+                cohort, interval_in_days
             )
-            # TODO: set current timestamp with the same timezone as the
-            # given clickstream and cohorts
 
             active_rates.name = name
             retention_rates = pd.concat([retention_rates, active_rates], axis=1)
